# Generated by Django 2.2.24 on 2021-07-07 17:46

from collections import defaultdict
from django.db import migrations


def create_repository_content_for_deprecation(apps, schema_editor):
    """ Create RepositoryContent for AnsibleCollectionDeprecated."""
    AnsibleCollectionDeprecated = apps.get_model('ansible', 'AnsibleCollectionDeprecated')
    RepositoryContent = apps.get_model('core', 'RepositoryContent')
    RepositoryVersionContentDetails = apps.get_model('core', 'RepositoryVersionContentDetails')
    db_alias = schema_editor.connection.alias
    repo_content = []
    counts_list = []
    count = defaultdict(int)

    for deprecation in AnsibleCollectionDeprecated.objects.using(db_alias).all():
        repo_content.append(RepositoryContent(
                content=deprecation,
                repository_id=deprecation.repository_id,
                version_added_id=deprecation.version_added_id,
            ))
        count[deprecation.version_added_id] += 1

    for repo_version_id, total in count.items():
        count_obj = RepositoryVersionContentDetails(
            content_type=deprecation.pulp_type,
            repository_version_id=repo_version_id,
            count=total,
            count_type="A", # Added
        )
        counts_list.append(count_obj)

        count_obj = RepositoryVersionContentDetails(
            content_type=deprecation.pulp_type,
            repository_version_id=repo_version_id,
            count=total,
            count_type="P", # Present
        )
        counts_list.append(count_obj)

    RepositoryContent.objects.using(db_alias).bulk_create(repo_content)
    RepositoryVersionContentDetails.objects.using(db_alias).bulk_create(counts_list)
    AnsibleCollectionDeprecated.objects.using(db_alias).filter(version_memberships__isnull=True).delete()


def create_deprecation_from_repository_content(apps, schema_editor):
    """ Create AnsibleCollectionDeprecated from RepositoryContent."""
    AnsibleCollectionDeprecated = apps.get_model('ansible', 'AnsibleCollectionDeprecated')
    Collection = apps.get_model('ansible', 'Collection')
    for deprecation in AnsibleCollectionDeprecated.objects.all():
        col = Collection.objects.get(namespace=deprecation.namespace, name=deprecation.name)
        for repo_content in deprecation.version_memberships.all():
            deprecation.collection_id = col.pk
            deprecation.repository_id = repo_content.repository_id
            deprecation.version_added_id = repo_content.version_added_id
            deprecation.save()


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0068_add_timestamp_of_interest'),
        ('ansible', '0035_deprecation_content'),
    ]

    operations = [
        migrations.RunPython(
            code=create_repository_content_for_deprecation,
            reverse_code=create_deprecation_from_repository_content,
            atomic=True,
        ),
        migrations.AlterUniqueTogether(
            name='ansiblecollectiondeprecated',
            unique_together={('namespace', 'name')},
        ),
        migrations.RemoveField(model_name="ansiblecollectiondeprecated", name="collection_id"),
        migrations.RemoveField(model_name="ansiblecollectiondeprecated", name="repository_id"),
        migrations.RemoveField(model_name="ansiblecollectiondeprecated", name="version_added_id"),
    ]



# coding=utf-8
"""Tests that CRUD remotes."""
from pulp_smash.pulp3.bindings import monitor_task

from pulp_ansible.tests.functional.utils import (
    gen_ansible_remote,
    TestCaseUsingBindings,
)
from pulp_ansible.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class CRUDAnsibleGitRemotesTestCase(TestCaseUsingBindings):
    """CRUD Ansible Git remotes."""

    def test_01_crud_remote(self):
        """Create, Read, Partial Update, and Delete an Ansible Git remote."""
        body = gen_ansible_remote(url="https://github.com/geerlingguy/ansible-role-adminer.git")
        remote = self.remote_git_api.create(body)
        self.addCleanup(self.remote_git_api.delete, remote.pulp_href)
        remote = self.remote_git_api.read(remote.pulp_href)
        for k, v in body.items():
            self.assertEquals(body[k], getattr(remote, k))
        update_body = {"url": "https://github.com/geerlingguy/ansible-role-ansible.git"}
        remote_update = self.remote_git_api.partial_update(remote.pulp_href, update_body)
        monitor_task(remote_update.task)
        remote = self.remote_git_api.read(remote.pulp_href)
        self.assertEqual(remote.url, update_body["url"])
        self.assertEqual(remote.metadata_only, False)
        with self.assertRaises(AttributeError):
            getattr(remote, "policy")

    def test_02_metadata_only_remote(self):
        """Create a remote where `metadata_only` is set to True."""
        body = gen_ansible_remote(
            url="https://github.com/geerlingguy/ansible-role-adminer.git", metadata_only=True
        )
        remote = self.remote_git_api.create(body)
        self.addCleanup(self.remote_git_api.delete, remote.pulp_href)
        for k, v in body.items():
            self.assertEqual(body[k], getattr(remote, k))



# Generated by Django 2.2.24 on 2021-07-07 17:46

from django.db import migrations, models
import django.db.models.deletion


def migrate_data_from_old_model_to_new_model_up(apps, schema_editor):
    """ Move objects from AnsibleCollectionDeprecated to NewAnsibleCollectionDeprecated."""
    AnsibleCollectionDeprecated = apps.get_model('ansible', 'AnsibleCollectionDeprecated')
    NewAnsibleCollectionDeprecated = apps.get_model('ansible', 'NewAnsibleCollectionDeprecated')
    AnsibleRepository = apps.get_model('ansible', 'AnsibleRepository')
    latest_repo_versions = []
    for repo in AnsibleRepository.objects.all():
        latest_repo_versions.append(repo.versions.exclude(complete=False).latest().pk)

    for deprecated_collection in AnsibleCollectionDeprecated.objects.filter(repository_version_id__in=latest_repo_versions):
        NewAnsibleCollectionDeprecated(
            namespace=deprecated_collection.collection.namespace,
            name=deprecated_collection.collection.name,
            pulp_type="ansible.collection_deprecation",
            collection_id=deprecated_collection.collection_id,
            repository_id=deprecated_collection.repository_version.repository.pk,
            version_added_id=deprecated_collection.repository_version.pk,
        ).save()


def migrate_data_from_old_model_to_new_model_down(apps, schema_editor):
    """ Move objects from NewAnsibleCollectionDeprecated to AnsibleCollectionDeprecated."""
    AnsibleCollectionDeprecated = apps.get_model('ansible', 'AnsibleCollectionDeprecated')
    NewAnsibleCollectionDeprecated = apps.get_model('ansible', 'NewAnsibleCollectionDeprecated')
    RepositoryVersionContentDetails = apps.get_model('core', 'RepositoryVersionContentDetails')
    for deprecation in NewAnsibleCollectionDeprecated.objects.all():
        AnsibleCollectionDeprecated(
            collection_id=deprecation.collection_id,
            repository_version_id=deprecation.version_added_id,
        ).save()
        deprecation.version_memberships.all().delete()
    RepositoryVersionContentDetails.objects.filter(content_type=deprecation.pulp_type).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0068_add_timestamp_of_interest'),
        ('ansible', '0034_handle_jsonfield_warnings'),
    ]

    operations = [
        migrations.CreateModel(
            name='NewAnsibleCollectionDeprecated',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_ansiblecollectiondeprecated', serialize=False, to='core.Content')),
                ('namespace', models.CharField(editable=False, max_length=64)),
                ('name', models.CharField(editable=False, max_length=64)),
                ('collection_id',  models.UUIDField(editable=False, primary_key=False, serialize=False, null=True)),
                ('repository_id',  models.UUIDField(editable=False, primary_key=False, serialize=False, null=True)),
                ('version_added_id',  models.UUIDField(editable=False, primary_key=False, serialize=False, null=True)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.content',),
        ),
        migrations.RunPython(
            code=migrate_data_from_old_model_to_new_model_up,
            reverse_code=migrate_data_from_old_model_to_new_model_down,
        ),
        migrations.DeleteModel(
            name='AnsibleCollectionDeprecated',
        ),
        migrations.RenameModel(
            old_name='NewAnsibleCollectionDeprecated',
            new_name='AnsibleCollectionDeprecated',
        ),
    ]



"""Tests collection sync functionality that is common to both Galaxy V2 and V3."""
from os import path
import subprocess
import tempfile

from pulp_ansible.tests.functional.utils import (
    gen_ansible_remote,
    SyncHelpersMixin,
    TestCaseUsingBindings,
)

from pulp_smash.pulp3.bindings import monitor_task
from pulp_smash.pulp3.utils import gen_distribution

from pulp_ansible.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class GitRemoteSyncInstallTestCase(TestCaseUsingBindings, SyncHelpersMixin):
    """Collection sync tests for collections with unique properties."""

    def test_sync_collection_from_git(self):
        """Sync collections from Git repositories and then install one of them."""
        body = gen_ansible_remote(url="https://github.com/pulp/pulp_installer.git")
        installer_remote = self.remote_git_api.create(body)
        self.addCleanup(self.remote_git_api.delete, installer_remote.pulp_href)

        repo = self._create_repo_and_sync_with_remote(installer_remote)

        content = self.cv_api.list(repository_version=repo.latest_version_href)
        self.assertEqual(len(content.results), 1)

        repo = self._sync_repo(repo, remote=installer_remote.pulp_href)

        content = self.cv_api.list(repository_version=repo.latest_version_href)
        self.assertEqual(len(content.results), 1)

        body = gen_ansible_remote(url="https://github.com/pulp/squeezer.git")

        squeezer_remote = self.remote_git_api.create(body)
        self.addCleanup(self.remote_git_api.delete, squeezer_remote.pulp_href)

        repo = self._sync_repo(repo, remote=squeezer_remote.pulp_href)

        content = self.cv_api.list(repository_version=repo.latest_version_href)
        self.assertEqual(len(content.results), 2)

        # Create a distribution.
        body = gen_distribution()
        body["repository"] = repo.pulp_href
        distribution_create = self.distributions_api.create(body)
        created_resources = monitor_task(distribution_create.task).created_resources
        distribution = self.distributions_api.read(created_resources[0])

        self.addCleanup(self.distributions_api.delete, distribution.pulp_href)
        collection_name = "pulp.squeezer"
        with tempfile.TemporaryDirectory() as temp_dir:

            # The install command needs --pre so a pre-release collection versions install
            cmd = "ansible-galaxy collection install --pre {} -c -s {} -p {}".format(
                collection_name, distribution.client_url, temp_dir
            )

            directory = "{}/ansible_collections/{}".format(
                temp_dir, collection_name.replace(".", "/")
            )

            self.assertTrue(
                not path.exists(directory), "Directory {} already exists".format(directory)
            )

            subprocess.run(cmd.split())

            self.assertTrue(path.exists(directory), "Could not find directory {}".format(directory))

    def test_sync_metadata_only_collection_from_git(self):
        """Sync collections from Git repositories with metadata_only=True."""
        body = gen_ansible_remote(
            url="https://github.com/ansible-collections/amazon.aws/",
            metadata_only=True,
            git_ref="2.1.0",
        )
        amazon_remote = self.remote_git_api.create(body)
        self.addCleanup(self.remote_git_api.delete, amazon_remote.pulp_href)

        repo = self._create_repo_and_sync_with_remote(amazon_remote)

        content = self.cv_api.list(repository_version=repo.latest_version_href)
        self.assertEqual(len(content.results), 1)

        # Create a distribution.
        body = gen_distribution()
        body["repository"] = repo.pulp_href
        distribution_create = self.distributions_api.create(body)
        created_resources = monitor_task(distribution_create.task).created_resources
        distribution = self.distributions_api.read(created_resources[0])

        self.addCleanup(self.distributions_api.delete, distribution.pulp_href)

        version = self.collections_versions_v3api.read(
            "aws", "amazon", distribution.base_path, "2.1.0"
        )

        self.assertEqual(version.git_url, "https://github.com/ansible-collections/amazon.aws/")
        self.assertEqual(version.git_commit_sha, "013162a952c7b2d11c7e2ebf443d8d4d7a21e95a")
        self.assertEqual(version.download_url, None)

    def test_sync_collection_from_git_commit_sha(self):
        """Sync collections from Git repositories with git_ref that is a commit sha."""
        body = gen_ansible_remote(
            url="https://github.com/ansible-collections/amazon.aws/",
            metadata_only=True,
            git_ref="d0b54fc082cb63f63d34246c8fe668e19e74777c",
        )
        amazon_remote = self.remote_git_api.create(body)
        self.addCleanup(self.remote_git_api.delete, amazon_remote.pulp_href)

        repo = self._create_repo_and_sync_with_remote(amazon_remote)

        content = self.cv_api.list(repository_version=repo.latest_version_href)
        self.assertEqual(len(content.results), 1)

        # Create a distribution.
        body = gen_distribution()
        body["repository"] = repo.pulp_href
        distribution_create = self.distributions_api.create(body)
        created_resources = monitor_task(distribution_create.task).created_resources
        distribution = self.distributions_api.read(created_resources[0])

        self.addCleanup(self.distributions_api.delete, distribution.pulp_href)

        version = self.collections_versions_v3api.read(
            "aws", "amazon", distribution.base_path, "1.5.1"
        )

        self.assertEqual(version.git_url, "https://github.com/ansible-collections/amazon.aws/")
        self.assertEqual(version.git_commit_sha, "d0b54fc082cb63f63d34246c8fe668e19e74777c")
        self.assertEqual(version.download_url, None)



import logging

from asgiref.sync import sync_to_async
from django.urls import reverse
from django.db.utils import IntegrityError
from galaxy_importer.collection import sync_collection
from gettext import gettext as _
from git import GitCommandError, Repo
from rest_framework.exceptions import ValidationError

from pulpcore.plugin.models import Artifact, ContentArtifact, ProgressReport
from pulpcore.plugin.stages import (
    DeclarativeArtifact,
    DeclarativeContent,
    DeclarativeVersion,
    Stage,
)
from pulp_ansible.app.models import AnsibleRepository, CollectionVersion, GitRemote
from pulp_ansible.app.tasks.collections import create_collection_from_importer

log = logging.getLogger(__name__)


def synchronize(remote_pk, repository_pk, mirror=False):
    """
    Sync content from the remote repository.

    Create a new version of the repository that is synchronized with the remote.

    Args:
        remote_pk (str): The remote PK.
        repository_pk (str): The repository PK.
        mirror (bool): True for mirror mode, False for additive.

    Raises:
        ValueError: If the remote does not specify a URL to sync.

    """
    remote = GitRemote.objects.get(pk=remote_pk)
    repository = AnsibleRepository.objects.get(pk=repository_pk)

    if not remote.url:
        raise ValueError(_("A remote must have a url specified to synchronize."))

    log.info(
        _("Synchronizing: repository=%(r)s remote=%(p)s"), {"r": repository.name, "p": remote.name}
    )
    first_stage = GitFirstStage(remote)
    d_version = DeclarativeVersion(first_stage, repository, mirror=mirror)
    return d_version.create()


class GitFirstStage(Stage):
    """
    The first stage of a pulp_ansible sync pipeline for git repositories.
    """

    def __init__(self, remote):
        """
        The first stage of a pulp_ansible sync pipeline.

        Args:
            remote (RoleRemote): The remote data to be used when syncing

        """
        super().__init__()
        self.remote = remote

        # Interpret download policy
        self.metadata_only = self.remote.metadata_only

    async def run(self):
        """
        Build and emit `DeclarativeContent` from the ansible metadata.
        """
        async with ProgressReport(
            message="Cloning Git repository for Collection", code="sync.git.clone"
        ) as pb:
            if self.remote.git_ref:
                try:
                    gitrepo = Repo.clone_from(
                        self.remote.url, self.remote.name, depth=1, branch=self.remote.git_ref
                    )
                except GitCommandError:
                    gitrepo = Repo.clone_from(self.remote.url, self.remote.name)
                    gitrepo.git.checkout(self.remote.git_ref)
            else:
                gitrepo = Repo.clone_from(self.remote.url, self.remote.name, depth=1)
            commit_sha = gitrepo.head.commit.hexsha
            metadata, artifact_path = sync_collection(gitrepo.working_dir, ".")
            if not self.remote.metadata_only:
                artifact = Artifact.init_and_validate(artifact_path)
                try:
                    await sync_to_async(artifact.save)()
                except IntegrityError:
                    artifact = Artifact.objects.get(sha256=artifact.sha256)
                metadata["artifact_url"] = reverse("artifacts-detail", args=[artifact.pk])
                metadata["artifact"] = artifact
            else:
                metadata["artifact"] = None
                metadata["artifact_url"] = None
            metadata["remote_artifact_url"] = "{}/commit/{}".format(
                self.remote.url.rstrip("/"), commit_sha
            )
            await self._add_collection_version(metadata)
            await pb.aincrement()

    async def _add_collection_version(self, metadata):
        """Add CollectionVersion to the sync pipeline."""
        artifact = metadata["artifact"]
        try:
            collection_version = await sync_to_async(create_collection_from_importer)(
                metadata, metadata_only=self.remote.metadata_only
            )
            await sync_to_async(ContentArtifact.objects.get_or_create)(
                artifact=artifact,
                content=collection_version,
                relative_path=collection_version.relative_path,
            )
        except ValidationError as e:
            if e.args[0]["non_field_errors"][0].code == "unique":
                namespace = metadata["metadata"]["namespace"]
                name = metadata["metadata"]["name"]
                version = metadata["metadata"]["version"]
            else:
                raise e
            collection_version = await sync_to_async(CollectionVersion.objects.get)(
                namespace=namespace, name=name, version=version
            )
        if artifact is None:
            artifact = Artifact()
        d_artifact = DeclarativeArtifact(
            artifact=artifact,
            url=metadata["remote_artifact_url"],
            relative_path=collection_version.relative_path,
            remote=self.remote,
            deferred_download=self.metadata_only,
        )

        # TODO: docs blob support??

        d_content = DeclarativeContent(
            content=collection_version,
            d_artifacts=[d_artifact],
        )
        await self.put(d_content)
